#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
%prog <FASTA> <BED>

Simulate DPs based on <FASTA> constrained by repeated regions <BED>.

Copyright (C) 2012-2016  Manuel Allhoff (allhoff@aices.rwth-aachen.de)

This program comes with ABSOLUTELY NO WARRANTY. This is free 
software, and you are welcome to redistribute it under certain 
conditions. Please see LICENSE file for details.
"""

from __future__ import print_function
from optparse import OptionParser
import sys
import pysam
import numpy as np
import HTSeq
from rgt.GenomicRegionSet import GenomicRegionSet
from rgt.GenomicRegion import GenomicRegion
from numpy.random import random_sample

from random import random, randint
from rgt.GenomicRegionSet import GenomicRegionSet
from rgt.GenomicRegion import GenomicRegion
from collections import Counter

MAX_FRAG_LEN = 500

class HelpfulOptionParser(OptionParser):
    """An OptionParser that prints full help on errors."""
    def error(self, msg):
        self.print_help(sys.stderr)
        self.exit(2, "\n%s: error: %s\n" % (self.get_prog_name(), msg))


def get_genome(ref_path, chrom):
    """Return genome for random access"""
    seq = None
    i = 0
    for s in HTSeq.FastaReader(ref_path):
        if s.name != chrom:
            i += 1
            continue
        else:
            seq = str(s)
            seq = seq.upper()
            break

    if i == 1:
        chrom = s.name
        seq = str(s)
    elif seq is None:
        parser.error("Sorry, the chromosome that is using for training (%s) is not contained \
        in the reference genome (%s)! Please use -c option!" %(chrom, ref_path))
    
    return seq, chrom

def sample_neg_bin(mean, dev):
    """Sample negative binomial for 1) number proteins and 2) number of fragments"""
    n = (mean*mean) / float(dev -mean)
    p = mean / float(dev)
    return np.random.negative_binomial(n, p) + 1

def get_left_protein_position(genome_seq, number_proteins, protein_space, valid_regions, bins):
    """Return position of left protein"""
    offset = (number_proteins - 1) * protein_space
    while True:
        valid = True
        pos = get_random_position(valid_regions, bins)

        #check whether pot. fragments have illegal nucleotide
        tmp_seq = genome_seq[pos - MAX_FRAG_LEN : pos + MAX_FRAG_LEN + offset]
        if tmp_seq.count('A') + tmp_seq.count('C') + tmp_seq.count('G') + tmp_seq.count('T') != len(tmp_seq):
            valid = False
        
        if valid:
            return pos

def get_frag_length(mean, dev):
    """Return length of fragment"""
    return min(int(np.random.normal(loc = mean, scale = dev)), MAX_FRAG_LEN)


def sample_fragments(protein_positions, genome_seq, frag_len_mean, frag_len_dev, frag_count_mean, frag_count_dev, dp_thres, min_counts, verbose, protein_space):
    """Sample fragments for all protein positions as list of (start, end, seq)"""
    fragments_sample1 = []
    fragments_sample2 = []
    peaks_sample1 = []
    peaks_sample2 = []
    
    for pos in protein_positions:
        counts_frag_sample1, counts_frag_sample2 = 0, 0
        number_frags = sample_neg_bin(frag_count_mean, frag_count_dev) 
        #read_frac = np.random.beta(0.5, 0.5) #fraction for sample 1
        read_frac = random()
        for _ in range(number_frags):
            frag_length = get_frag_length(frag_len_mean, frag_len_dev)
            qpos = pos + randint(-protein_space/3, protein_space/3) #randomly shift fragement around protein
            seq = genome_seq[qpos - frag_length/2 : qpos + frag_length/2]
            
            if random() <= read_frac:
                fragments_sample1.append((pos - frag_length/2, pos + frag_length/2, seq))
                counts_frag_sample1 += 1
            else:
                fragments_sample2.append((pos - frag_length/2, pos + frag_length/2, seq))
                counts_frag_sample2 += 1
        
        if counts_frag_sample1 > min_counts and \
        (counts_frag_sample1 / float(counts_frag_sample1 + counts_frag_sample2) >= dp_thres or counts_frag_sample1 - counts_frag_sample2 > 1.5*min_counts):
            peaks_sample1.append(pos)
        
        if counts_frag_sample2 > min_counts and \
        (counts_frag_sample2 / float(counts_frag_sample1 + counts_frag_sample2) >= dp_thres or counts_frag_sample2 - counts_frag_sample1 > 1.5*min_counts):
            peaks_sample2.append(pos)              
        
    return fragments_sample1, fragments_sample2, peaks_sample1, peaks_sample2

def reverse_complement(s, rev=True):
    """Return the reverse complement of a DNA sequence s"""
    _complement = dict(A="T", T="A", C="G", G="C", N="N")
    t = reversed(s) if rev else s
    try:
        rc = (_complement[x] for x in t)  # lazy generator expression
    except KeyError:
        return s
        
    return "".join(rc)  # join the elements into a string

def write_bed(chrom, peaks, name):
    """Write bedfiles with peaks."""
    bedfile = open(name, "w")
    for p in peaks:
        bedfile.write("%s\t%s\t%s\n" %(chrom, p, p+1))
    
def write_fasta(fragments, read_length, name):
    """Write reads from both end of fragments"""
    fastafile = open(name, "w")
    i = 0
    for _, _, seq in fragments:
        seq_forward = seq[:read_length]
        seq_reverse = seq[len(seq) - read_length:]
        if not len(seq_forward) == len(seq_reverse):
            seq_forward = seq
            seq_reverse = seq
        seq_reverse = reverse_complement(seq_reverse)
        
        fastafile.write(">read%s\n%s\n" %(i, seq_forward))
        i += 1
        fastafile.write(">read%s\n%s\n" %(i, seq_reverse))
        i += 1
        

def get_valid_regions(befile_path, genome_length):
    repeat_regions = GenomicRegionSet('repeat')
    valid_regions = GenomicRegionSet('valid')
    
    repeat_regions.read_bed(bedfile_path)
    valid_regions.add(GenomicRegion(options.chrom, 0, genome_length))
    valid_regions = valid_regions.subtract(repeat_regions)
    s = 0
    weights = []
    for el in valid_regions:
        s += len(el)
        weights.append(len(el))
    prob = map(lambda x: x/float(s), weights)
    bins = np.add.accumulate(prob)
    
    return valid_regions, bins

def get_random_position(valid_regions, bins):
    chosen_region = np.digitize(random_sample(1), bins)
    return randint(valid_regions.sequences[chosen_region].initial, valid_regions.sequences[chosen_region].final)


if __name__ == '__main__':
    test = False
    
    if not test:
        parser = HelpfulOptionParser(usage=__doc__)
        
        parser.add_option("-c", "--chrom", default='chr1', dest="chrom", type="string", help="Set chromosome to simulate. [default: %default]")
        parser.add_option("-s", "--space", default=200, dest="protein_space", type="int", help="Set space between to proteins. [default: %default]")
        parser.add_option("-m", "--min-counts", default=25, dest="min_counts", type="int", help="Set minimum number of reads for a DP. [default: %default]")
        parser.add_option("-v", "--verbose", default=False, dest="verbose", action="store_true", help="Output info file. [default: %default]")
        parser.add_option("-l", "--length", default=26, dest="read_length", type="int", help="Set read's length. [default: %default]")
        parser.add_option("-p", "--prefix", default='simulate', dest="prefix", type="string", help="Set prefix of output files. [default: %default]")
        parser.add_option("-n", "--peak-counts", default=2000, dest="peak_counts", type="int", help="Set number of regions with DPs. [default: %default]")
         
        parser.add_option("--dp-thres", default=0.6, dest="dp_thres", type="float", help="Set read fraction to define DP. [default: %default]")
         
        parser.add_option("--frag-len-mean", default=200, dest="frag_len_mean", type="float", help="Set mean of fragments. [default: %default]")
        parser.add_option("--frag-len-dev", default=20, dest="frag_len_dev", type="float", help="Set deviation of fragments. [default: %default]")
         
        parser.add_option("--frag-count-mean", default=100, dest="frag_count_mean", type="float", help="Set parameter n of neg. bin. distribution for fragment counts. [default: %default]")
        parser.add_option("--frag-count-dev", default=200, dest="frag_count_dev", type="float", help="Set parameter p of neg. bin. distribution for fragment counts. [default: %default]")
         
        parser.add_option("--prot-count-mean", default=1, dest="prot_count_mean", type="float", help="Set parameter n of neg. bin. distribution for fragment counts. [default: %default]")
        parser.add_option("--prot-count-dev", default=4, dest="prot_count_dev", type="float", help="Set parameter p of neg. bin. distribution for fragment counts. [default: %default]")
        
        
        (options, args) = parser.parse_args()
        
        i = 2
        if len(args) != i:
            parser.error("Exactly %s parameters are needed" %i)
         
        fastafile_path = args[0]
        bedfile_path = args[1]
    else:
        parser = HelpfulOptionParser(usage=__doc__)
        (options, args) = parser.parse_args()
        fastafile_path = '/home/manuel/test_sim/test_chrom.fasta'
        bedfile_path = '/home/manuel/test_sim/test_repeat.bed'
       
        options.chrom = 'chr1'
        options.protein_space = 180
        options.dp_thres = 0.6
        options.min_counts = 25
        options.verbose = True
        options.read_length = 26
        options.prefix = "test"
        options.peak_counts = 2000
         
        options.frag_len_mean = 200
        options.frag_len_dev = 20
         
        options.frag_count_mean = 100
        options.frag_count_dev = 2000
         
        options.prot_count_mean = 1
        options.prot_count_dev = 4
    
    all_fragments_sample1 = []
    all_fragments_sample2 = []
    all_peaks_sample1 = []
    all_peaks_sample2 = []
    
    genome_seq, _ = get_genome(fastafile_path, options.chrom)
    valid_regions, bins = get_valid_regions(bedfile_path, len(genome_seq))
    valid_regions.extend(MAX_FRAG_LEN, MAX_FRAG_LEN)
    
    j = 0
    for _ in range(options.peak_counts):
        j += 1
#         print(j, file=sys.stderr)
        number_proteins = sample_neg_bin(options.prot_count_mean, options.prot_count_dev)
        protein_left_position = get_left_protein_position(genome_seq, number_proteins, options.protein_space, valid_regions, bins)
        protein_positions = [protein_left_position + i * options.protein_space for i in range(number_proteins)]
         
         
        fragments_sample1, fragments_sample2, peaks_sample1, peaks_sample2 = \
        sample_fragments(protein_positions, genome_seq, options.frag_len_mean, options.frag_len_dev, options.frag_count_mean, options.frag_count_dev, options.dp_thres, options.min_counts, options.verbose, options.protein_space)      
         
        all_fragments_sample1 += fragments_sample1
        all_fragments_sample2 += fragments_sample2
        all_peaks_sample1 += peaks_sample1
        all_peaks_sample2 += peaks_sample2
     
    write_fasta(all_fragments_sample1, options.read_length, options.prefix + "sample1.fasta")
    write_fasta(all_fragments_sample2, options.read_length, options.prefix + "sample2.fasta")
     
    write_bed(options.chrom, all_peaks_sample1, options.prefix + "sample1-peaks.bed")
    write_bed(options.chrom, all_peaks_sample2, options.prefix + "sample2-peaks.bed")
    
    
    
    
    
